#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdbool.h>
#include <assert.h>

#include "vector.h"

struct array {
	size_t count; // currently used slots
	size_t slots; // total slots
	size_t size; // size of a slot
	void *arr; // the actual array
};

static inline bool resize(struct array *arr, size_t new_size)
{
	assert(arr->count < new_size);
	arr->arr = realloc(arr->arr, new_size * arr->size);
	if(!arr->arr)
		return false;
	arr->slots = new_size;
	return true;
}

void *init_vector(size_t nmemb, size_t size)
{
	struct array *ret;
	void *p;

	// Check for overflow
	if (size && nmemb > (size_t)-1/size) {
		errno = ENOMEM;
		return 0;
	}

	p = malloc(size * nmemb);
	if(!p)
		return p;
	memset(p, 0, size);

	ret = malloc(sizeof(struct array));
	if(!ret) {
		free(p);
		return ret;
	}

	ret->count = 0;
	ret->slots = nmemb;
	ret->size = size;
	ret->arr = p;

	return ret;
}

void free_vector(struct array *arr)
{
	free(arr->arr);
	free(arr);
}

bool push_back(struct array *arr, void *el)
{
	return insert_at(arr, arr->count, el);
}

bool insert_at(struct array *arr, size_t pos, void *el)
{
	char *from, *to;
	size_t size;

	if(pos > arr->count)
		return false;

	if(arr->count + 1 > arr->slots)
		if(!resize(arr, arr->slots * 2))
			return false;

	from = (char *)arr->arr + pos * arr->size;
	if(pos + 1 < arr->slots) {
		to = from + arr->size;
		size = (arr->count - pos) * arr->size;
		memmove(to, from, size);
	}

	memcpy(from, el, arr->size);

	arr->count++;
	assert(arr->count <= arr->slots);
	return true;
}

bool read_from(struct array *arr, size_t pos, void *buffer)
{
	if(pos < arr->count) {
		memcpy(buffer, (char *)arr->arr + pos * arr->size, arr->size);
		return true;
	}

	return false;
}

bool extract_from(struct array *arr, size_t pos, void *buffer)
{
	char *from, *to;
	size_t size;

	if(!read_from(arr, pos, buffer))
		return false;

	if(pos + 1 < arr->slots) {
		from = (char *)arr->arr + (pos + 1) * arr->size;
		to = from - arr->size;
		size = (arr->count - pos) * arr->size;
		memmove(to, from, size);
	}

	arr->count--;
	if(arr->count < arr->slots / 2)
		resize(arr, arr->slots / 2);
	return true;
}

